
--- Simple phantom type example
--- Safe Lists

module SList (Unsafe, Safe, List, nil, cons, sHead, sTail, sInit, sLast, sElem, sMap, sFilter, testnull, forget, toList, fromList) where

import List

-- The phantom type 
data List a b = L [b]

instance Show b => Show (List a b) where
   show (L l) = show l
	 
-- known to be non-empty
data Safe = Safe
-- may be empty
data Unsafe = Unsafe


-- Constructors and operations typed with the phantom type
-- are just wrappers for standard list versions

nil :: List Unsafe a
nil = (L [])

cons :: b -> List a b -> List Safe b
cons hd (L tl) = L (hd : tl)

-- a more informative version of 'null'

testnull :: List a b -> (List Unsafe b -> c) -> (List Safe b -> c) -> c
testnull (L l) isnull nonnull = 
    if null l then isnull (L l) else nonnull (L l)


-- For flexibilty, note we don't want the opposite conversion

forget :: List Safe b -> List Unsafe b
forget (L x) = (L x)

toList :: List a b -> [b]
toList (L x) = x 

fromList :: [b] -> List Unsafe b
fromList = L

-- Four operations that require non-empty lists

sHead :: List Safe a -> a 
sHead (L l) = head l

sTail :: List Safe a -> List Unsafe a
sTail (L l) = L (tail l)

sLast :: List Safe a -> a
sLast (L l) = last l

sInit :: List Safe a -> List Unsafe a 
sInit (L l) = L (init l)

-- Other standard operations that work for any list

sElem :: Eq b => b -> List a b -> Bool
sElem b (L l) = elem b l

sMap  :: (b1 -> b2) -> List a b1 -> List a b2
sMap f (L l) = L (map f l)

sFilter :: (b -> Bool) -> List a b -> List Unsafe b
sFilter f (L l) = L (filter f l)

sLength :: List a b -> Int
sLength (L l) = length l

-- what is type of (binary) concat??
sConcat :: List a1 b -> List a2 b -> List Unsafe b
sConcat (L l1) (L l2) = L (l1 ++ l2)



